/*
 * asteriskaudio.cxx
 *
 * Sound driver implementation for ASTERISK PBX.
 * OpenH323 Wrapper Library.
 *
 * Copyright (c) 2002,2003 InAccess Networks
 * Michalis Manousos <manousos@inaccessnetworks.com>
 * Dimitris Economou <decon@inaccessnetworks.com>
 *
 * This file is part of "H323 support for ASTERISK"
 *
 * "H323 support for ASTERISK" is free software;
 * you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation;
 * either version 2 of the License, or (at your option) any later version. 
 *
 * "H323 support for ASTERISK" is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details. 
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
 *
 * $Id: asteriskaudio.cxx,v 1.21 2003/08/07 16:17:10 manousos Exp $
 *
 */

#include <ptlib.h>
#include <sys/time.h>
#include <rtp.h>

#include "oh323.h"
#include "version.h"
#include "asteriskaudio.hxx"


/***********************************************************
 * Class PAsteriskAudioDelay
 **********************************************************/

PAsteriskAudioDelay::PAsteriskAudioDelay()
{
	WRAPTRACE(1, "Object initialized.");

	firstTime = TRUE;
	error = 0;
}
    
PAsteriskAudioDelay::~PAsteriskAudioDelay()
{
	WRAPTRACE(1, "Object deleted.");
}
    
void PAsteriskAudioDelay::Restart()
{
	WRAPTRACE(2, "Restart.");

	firstTime = TRUE;
	//error = 0;
}
  
BOOL PAsteriskAudioDelay::ReadDelay(int frameTime)
{
  if (firstTime) {
    firstTime = FALSE;
    targetTime = PTime();   // targetTime is the time we want to delay to
    return TRUE;
  }

  // Set the new target
  targetTime += frameTime;

  // Calculate the sleep time so we delay until the target time
  PTimeInterval delay = targetTime - PTime();
  int sleep_time = (int)delay.GetMilliSeconds();

  if (sleep_time > 0)
    usleep(sleep_time * 1000);

  return sleep_time <= -frameTime;
}
BOOL PAsteriskAudioDelay::WriteDelay(int frameTime)
{
	if (firstTime) {
		firstTime = FALSE;
		targetTime = PTime();
		error = 0;
		return TRUE;
	}

	targetTime += frameTime;
	targetTime += error;		// Decrease it

	PTime now;
	PTimeInterval delay = targetTime - now;
	int sleep_time = (int)delay.GetMilliSeconds();
	if (sleep_time <= 0) {
		WRAPTRACE(2, "Schedule in the past???");
		error = sleep_time;
	} else {
		error = 0;
		usleep(sleep_time * 1000);
	}

	return error <= -frameTime;
}
BOOL PAsteriskAudioDelay::Delay(int frameTime)
{
	if (firstTime) {
		firstTime = FALSE;
		previousTime = PTime();
		return TRUE;
	}

	error += frameTime;

	PTime now;
	PTimeInterval delay = now - previousTime;
	error -= (int)delay.GetMilliSeconds();
	previousTime = now;

	if (error > 0)
		usleep(error * 1000);

	return error <= -frameTime;
}

/***********************************************************
 * Class PAsteriskSoundChannel
 **********************************************************/

PAsteriskSoundChannel::PAsteriskSoundChannel()
{
	WRAPTRACE(1, "Object initialized.");
	mutex_flag = 0;
	lastWritten = 0;
	frameTime = 0;
	frameNum = 0;
	time_diff = 0;
	first_write = TRUE;
	//SetReadTimeout(0);
	//astWriteDelay.Delay(0);
	//astWriteDelay.ReadDelay(0);
	astReadDelay.ReadDelay(0);

	Construct();
}

void PAsteriskSoundChannel::Construct()
{
}

PAsteriskSoundChannel::~PAsteriskSoundChannel()
{
	Close();
	WRAPTRACE(1, "Object deleted.");
}

PStringArray PAsteriskSoundChannel::GetDeviceNames(Directions dir)
{
	PStringArray	recordArray, playArray;
	PStringArray	error;
	int				i;
	char			buf[MAX_FIFO_NAME];

	error[0] = "";

	/* Fill the recordArray */
	for (i=0; i<AST_OH323_MAX_CALLS; i++) {
		memset(buf, 0, MAX_FIFO_NAME);
		snprintf(buf, MAX_FIFO_NAME-1, "%sin%d", AST_OH323_FIFO_PREFIX, i);
		recordArray[i] = buf;
	}

	/* Fill the playArray */
	for (i=0; i<AST_OH323_MAX_CALLS; i++) {
		memset(buf, 0, MAX_FIFO_NAME);
		snprintf(buf, MAX_FIFO_NAME-1, "%sout%d", AST_OH323_FIFO_PREFIX, i);
		playArray[i] = buf;
	}

	if (dir == Recorder) {
		return recordArray;
	} else if (dir == Player) {
		return playArray;
	} else {
		return error;
	}
}

PString PAsteriskSoundChannel::GetDefaultDevice(Directions dir)
{
	PString		res = AST_OH323_FIFO_PREFIX;

	if (dir == Recorder) {
		res += "in0";
		return res;
	} else if (dir == Player) {
		res += "out0";
		return res;
	} else {
		return "";
	}
}

BOOL PAsteriskSoundChannel::Open(const PString & device,
								 int deviceFd,
								 Directions dir,
								 unsigned numChannels,
								 unsigned mediaFmt,
								 unsigned frameTm,
								 int frameNm)
{
	Close();

	/***********
	if (!ConvertOSError(os_handle = ::open(device, dir == Player ? O_WRONLY : O_RDONLY | O_NONBLOCK)))
		return FALSE;
	**********/
	if (deviceFd < 0)
		return FALSE;
	os_handle = deviceFd;
	mediaFormat = mediaFmt;
	frameTime = frameTm;
	frameNum = frameNm;
	WRAPTRACE(2, "os_handle " << os_handle << ", mediaFormat " << mediaFormat
				<< ", frameTime " << frameTime << ", frameNum " << frameNum);
	return TRUE;
}

BOOL PAsteriskSoundChannel::Close()
{
	// XXX Do not close it from in here, do it from the channel driver.
	os_handle = -1;
	return TRUE;
#if 0
	WRAPTRACE(2, "Closing os_handle " << os_handle);
	if (PChannel::Close() != TRUE) {
		// WRAPTRACE(10, "Failed - " << GetErrorText());
		return FALSE;
	} else
		return TRUE;
#endif
}

BOOL PAsteriskSoundChannel::SetFormat(unsigned numChannels,
							  unsigned sampleRate,
							  unsigned bitsPerSample)
{
	Abort();
	PAssert(numChannels >= 1 && numChannels <= 2, PInvalidParameter);
	PAssert(bitsPerSample == 8 || bitsPerSample == 16, PInvalidParameter);
	return TRUE;
}

BOOL PAsteriskSoundChannel::SetBuffers(PINDEX size, PINDEX count)
{
	Abort();
	PAssert(size > 0 && count > 0 && count < 65536, PInvalidParameter);
	bufSize = size;
	bufCount = count;

	return TRUE;
}

BOOL PAsteriskSoundChannel::GetBuffers(PINDEX & size, PINDEX & count)
{
	size = bufSize;
	count = bufCount;

	return TRUE;
}
#if 0
BOOL PAsteriskSoundChannel::Write(const void * buffer, PINDEX length)
{
	//WRAPTRACE(10, "In Write... " << length);
	if (PChannel::Write(buffer, length) != TRUE) {
		WRAPTRACE(3, "Failed - " << GetErrorText());
		return FALSE;
	} else {
		// Do a sleep to simulate the amount of time the hardware would take
		// to write out this data.
		switch (mediaFormat) {
			//case RTP_DataFrame::L16_Mono:
				// 1 msec/frame
				// 1 frame=16 bytes
			//	astWriteDelay.Delay(length/16);
			//	break;
			case RTP_DataFrame::PCMU:
			case RTP_DataFrame::PCMA:
				//astWriteDelay.Delay(length/16);
				astWriteDelay.Delay(length/8);
				break;
			case RTP_DataFrame::G729:
				// 10 msec/frame
				astWriteDelay.Delay(10);
				break;
			case RTP_DataFrame::GSM:
				// 20 msec/frame
				astWriteDelay.Delay(20);
				break;
			case RTP_DataFrame::G7231:
				// 30 msec/frame
				astWriteDelay.Delay(30);
				break;
			default:
				WRAPTRACE(3, "Unknown media format " << mediaFormat);
				break;
		}

		return TRUE;
	}
}
#endif

static unsigned const G7231FrameSize[4] = { 24, 20, 4, 1};
BOOL PAsteriskSoundChannel::Write(const void * buffer, PINDEX length)
{
	int written, size, num;
	char *buf;

	if (length == 0)
		WRAPTRACE(2, "Request to write 0 bytes!");

	if (os_handle < 0) {
		WRAPTRACE(2, "Channel is shutting down!");
		return FALSE;
	}

/************
	if (first_write == TRUE) {
		PTime time1;
		previous_write = time1;
		time_diff = 0;
		first_write = FALSE;
	}
************/
	// Do a sleep to simulate the amount of time the hardware would take
	// to write out this data.
	lastWritten = 0;
	switch (mediaFormat) {
		//case RTP_DataFrame::L16_Mono:
			// 1 msec/frame
			// 1 frame=16 bytes
		//	astWriteDelay.Delay(length/16);
		//	break;
		case RTP_DataFrame::PCMU:
		case RTP_DataFrame::PCMA:
			if (PChannel::Write(buffer, length) != TRUE) {
				WRAPTRACE(2, "Write failed (G.711) - " << GetErrorText());
				return FALSE;
			} else {
				// 1 msec/frame
				num = length/8;
				astWriteDelay.Delay(num * frameTime);
				//astWriteDelay.WriteDelay(num * frameTime);
				lastWritten = length;
			}
			break;
		case RTP_DataFrame::G729:
			/* XXX Not complete... */
			if (PChannel::Write(buffer, length) != TRUE) {
				WRAPTRACE(2, "Write failed (G.729) - " << GetErrorText());
				return FALSE;
			} else {
				// 10 msec/frame
				astWriteDelay.Delay(frameTime);
				lastWritten = length;
			}
			break;
		case RTP_DataFrame::GSM:
			if (PChannel::Write(buffer, length) != TRUE) {
				WRAPTRACE(2, "Write failed (GSM) - " << GetErrorText());
				return FALSE;
			} else {
				// 20 msec/frame
				num = length/33;
				astWriteDelay.Delay(num * frameTime);
				lastWritten = length;
			}
			break;
		case RTP_DataFrame::G7231:
#if 0
			//WRAPTRACE(10, "Length is " << length);
			if (PChannel::Write(buffer, length) != TRUE) {
				WRAPTRACE(3, "Write failed (G.723.1) - " << GetErrorText());
				return FALSE;
			} else {
				// 30 msec/frame
				astWriteDelay.Delay(30);
				lastWritten = length;
			}
			break;
#endif
			buf = (char*)buffer;
			size = length;
			while (size > 0) {
				written = G7231FrameSize[buf[length-size]&3];
				if (written != 1) {
					if (PChannel::Write((const void *)&buf[length-size], written) != TRUE) {
						WRAPTRACE(2, "Write failed (G.723.1) - " << GetErrorText());
						return FALSE;
					} else {
						// 30 msec/frame
						astWriteDelay.Delay(frameTime);
					}
				} else {	// written == 1 (Silence)
					// 30 msec/frame
					astWriteDelay.Delay(frameTime);
				}
				size -= written;
				lastWritten += written;
			}
			break;
		default:
			WRAPTRACE(2, "Unknown media format " << mediaFormat);
			break;
	}

/*************
	PTime time2;
	time_diff = time2 - previous_write;
	previous_write = time2;
	WRAPTRACE(2, "Written [" << GetLastWriteCount() <<" bytes] [" << 
			time_diff.GetMilliSeconds() << " ms]");
************/
	WRAPTRACE(5, "Written [" << GetLastWriteCount() <<" bytes]");
	return TRUE;
}

PINDEX PAsteriskSoundChannel::GetLastWriteCount() const
{
	return(lastWritten < 0 ? 0 : lastWritten);
}

BOOL PAsteriskSoundChannel::Read(void * buffer, PINDEX length)
{
	int readTimeout = -1;
	lastReadCount = 0;
	
	if (os_handle < 0) {
		WRAPTRACE(2, "Channel is shutting down!");
		return FALSE;
	}

	switch (mediaFormat) {
		case RTP_DataFrame::PCMU:
		case RTP_DataFrame::PCMA:
			// 1 msec/frame * txFrames
			readTimeout = frameTime * (length/8);
			break;
		case RTP_DataFrame::G729:
			// 10 msec/frame
		case RTP_DataFrame::GSM:
			// 20 msec/frame
		case RTP_DataFrame::G7231:
			// 30 msec/frame
			readTimeout = frameTime;
			//SetReadTimeout(frameTime);
			break;
		default:
			WRAPTRACE(2, "Unknown media format " << mediaFormat);
			break;
	}

	/************
	if (readTimeout > 0) {
		if (astReadDelay.Delay(readTimeout) != TRUE) {
			WRAPTRACE(2, "Delay returned true!");
			astReadDelay.Restart();
		}
	} else
		return FALSE;
	************/
	if (readTimeout > 0) {
		astReadDelay.ReadDelay(readTimeout);
	} else
		return FALSE;
	SetReadTimeout(0);
	if (PChannel::Read(buffer, length) != TRUE) {
		if (GetErrorCode(LastReadError) == PChannel::Timeout) {
			WRAPTRACE(4, "Timeout [" << GetLastReadCount() << " bytes]");
		} else if (GetErrorCode(LastReadError) != PChannel::Interrupted) {
			WRAPTRACE(2, "Failure - " << GetErrorText());
			return FALSE;
		}
	} else {
		WRAPTRACE(5, "Data read [" << GetLastReadCount() << " bytes]");
	}
	return TRUE;
}


#if 0
BOOL PAsteriskSoundChannel::Read(void * buffer, PINDEX length)
{
	/* NOTE: See codecs.cxx/h, lid.cxx/h, channels.cxx of OpenH323
	         to understand how this function (Read) should work.
	 */
	int readTimeout = 0;

	lastReadCount = 0;

	switch (mediaFormat) {
		case RTP_DataFrame::PCMU:
		case RTP_DataFrame::PCMA:
			// 1 msec/frame * txFrames
			readTimeout = frameTime * frameNum;
			//SetReadTimeout(frameTime * frameNum);
			break;
		case RTP_DataFrame::G729:
			// 10 msec/frame
		case RTP_DataFrame::GSM:
			// 20 msec/frame
		case RTP_DataFrame::G7231:
			// 30 msec/frame
			readTimeout = frameTime;
			//SetReadTimeout(frameTime);
			break;
		default:
			WRAPTRACE(2, "Unknown media format " << mediaFormat);
			//cout << "Error: Unknown media format " << mediaFormat << endl;
			break;
	}
	if (readTimeout > 0)
		//SetReadTimeout(0);
		//SetReadTimeout(readTimeout);
		;
	else
		return FALSE;

	//PTime time1;
	if (PChannel::Read(buffer, length) != TRUE) {
		if (GetErrorCode(LastReadError) == PChannel::Timeout) {
			astReadDelay.Restart();
			//astReadDelay.Delay(readTimeout);
			WRAPTRACE(5, "Timeout while waiting for frame in the last " <<
				GetReadTimeout().GetInterval() << " ms.");
		} else {
			WRAPTRACE(2, "Failed - " << GetErrorText());
			return FALSE;
		}
	} else {
		//PTime time2;
		//PTimeInterval readDT = time2 - time1;
		//int dtime = readTimeout - (int)readDT.GetMilliSeconds();	// ms
		//if (dtime > 0)
		////	//usleep(dtime * 1000);
		//	astReadDelay.Delay(dtime);
		astReadDelay.Delay(readTimeout);
		WRAPTRACE(5, "Read [" << GetLastReadCount() <<" bytes]");
	}

	/* Maintain the real-time nature of this function */
//	if (astReadDelay.Delay(readTimeout) != TRUE) {
//	}
	return TRUE;
}
#endif
// End of file /////////////////////////////////////////////////////////////
